{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "0fe7ca78-244b-4119-87d6-9cd67b07adfd",
   "metadata": {},
   "source": [
    "## 必要なライブラリのimport(Notebookを開いたり, Kernel再起動の度に一度だけ実行)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "78224e39-380c-4653-b2d2-e43465fdc015",
   "metadata": {},
   "outputs": [],
   "source": [
    "import shutil, os\n",
    "import pandas as pd\n",
    "import numpy as np\n",
    "import rclpy\n",
    "from rclpy.node import Node\n",
    "from geometry_msgs.msg import TransformStamped\n",
    "import copy\n",
    "from std_msgs.msg import Empty\n",
    "from triorb_drive_interface.msg import TriorbSetPos3 as TriorbSetPos3Topic\n",
    "from triorb_drive_interface.msg import TriorbRunResult\n",
    "from triorb_drive_interface.msg import DriveGains\n",
    "from scipy.spatial.transform import Rotation as R\n",
    "rclpy.init()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6bd65e78-9447-41e1-a0e1-3993889be241",
   "metadata": {},
   "source": [
    "## 事前定義(Notebookを開いたり, Kernel再起動の度に一度だけ実行)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "bab0d6ab-9e58-4315-9808-16f5699131e0",
   "metadata": {},
   "outputs": [],
   "source": [
    "STD_ACC = 1000\n",
    "STD_DEC = 1000\n",
    "GAIN_XY_P = 0.22 # 0.2\n",
    "GAIN_XY_I = 0.0\n",
    "GAIN_XY_D = 1e-4 # -5\n",
    "GAIN_W_P = 0.20\n",
    "GAIN_W_I = 0.0\n",
    "GAIN_W_D = 1e-5 # -6\n",
    "\n",
    "def get_topic_name(topic_name_raw):\n",
    "    return (os.getenv('ROS_PREFIX', '')+'/'+topic_name_raw).replace(\"//\",\"/\")\n",
    "\n",
    "def quaternion_to_euler_zyx(q):\n",
    "    r = R.from_quat([q[0], q[1], q[2], q[3]])\n",
    "    return r.as_euler('zyx', degrees=False)\n",
    "    #return r.as_euler('zyx', degrees=True)\n",
    "\n",
    "class Pose:\n",
    "    def __init__(self, x,y,z,roll,pitch,yaw):\n",
    "        self.x = x\n",
    "        self.y = y\n",
    "        self.z = z\n",
    "        self.roll = roll\n",
    "        self.pitch = pitch\n",
    "        self.yaw = yaw\n",
    "        self.deg = -np.rad2deg(yaw)\n",
    "\n",
    "class LocalizeNode(Node):\n",
    "    def __init__(self):\n",
    "        super().__init__('localize_node')\n",
    "        self.sub_tf = self.create_subscription(TransformStamped, get_topic_name('/vslam/rig_tf'), self.callback_rig_tf, 10)\n",
    "\n",
    "    def destroy(self):\n",
    "        _ = self.destroy_subscription(self.sub_tf)\n",
    "        self.destroy_node()\n",
    "    \n",
    "    def callback_rig_tf(self, msg):\n",
    "        _x, _y, _z = msg.transform.translation.x, msg.transform.translation.y, msg.transform.translation.z\n",
    "        _yaw, _pitch, _roll  = quaternion_to_euler_zyx([msg.transform.rotation.x, msg.transform.rotation.y, msg.transform.rotation.z, msg.transform.rotation.w])\n",
    "        self.pose = Pose( _x, _y, _z, _roll, _pitch, _yaw )\n",
    "        raise KeyboardInterrupt\n",
    "\n",
    "\n",
    "class NavigateNode(Node):\n",
    "    def __init__(self):\n",
    "        super().__init__('sample_abs_pos')\n",
    "        self.pub_gain = self.create_publisher(DriveGains, get_topic_name(\"/setting/drive/gains\"), 1)\n",
    "        self.pub_set_pos = self.create_publisher(TriorbSetPos3Topic, get_topic_name(\"/drive/set_pos\"), 1)\n",
    "        self.pub_stop = self.create_publisher(Empty, get_topic_name(\"/drive/stop\"), 10)\n",
    "        self.subs = [self.create_subscription(TriorbRunResult, get_topic_name('/drive/result'), self.callback_result, 1)]\n",
    "\n",
    "    def run_pos(self, x,y,deg, vxy=0.2, vw=0.5, tx=0.1, ty=0.1, tr=5.0, force=1):\n",
    "        request = TriorbSetPos3Topic()\n",
    "        request.pos.position.x = x\n",
    "        request.pos.position.y = y\n",
    "        request.pos.position.deg = deg\n",
    "        request.pos.speed.xy = vxy\n",
    "        request.pos.speed.w = vw\n",
    "        request.pos.speed.acc = STD_ACC\n",
    "        request.pos.speed.dec = STD_DEC\n",
    "        request.setting.tx = tx\n",
    "        request.setting.ty = ty\n",
    "        request.setting.tr = tr\n",
    "        request.setting.force = force\n",
    "        self.pub_set_pos.publish( request )\n",
    "        raise KeyboardInterrupt\n",
    "            \n",
    "    def set_gain(self):\n",
    "        msg = DriveGains()\n",
    "        msg.xy_p = GAIN_XY_P\n",
    "        msg.xy_i = GAIN_XY_I\n",
    "        msg.xy_d = GAIN_XY_D\n",
    "        msg.w_p = GAIN_W_P\n",
    "        msg.w_i = GAIN_W_I\n",
    "        msg.w_d = GAIN_W_D\n",
    "        self.pub_gain.publish(msg)\n",
    "        raise KeyboardInterrupt\n",
    "\n",
    "    def callback_result(self, msg):\n",
    "        print(\"Receive results\")\n",
    "        self.get_logger().info(str(msg))\n",
    "        raise KeyboardInterrupt\n",
    "\n",
    "    def stop(self):\n",
    "        self.pub_stop.publish(Empty())\n",
    "        raise KeyboardInterrupt\n",
    "    \n",
    "    def destroy(self):\n",
    "        self.pub_stop.publish(Empty())\n",
    "        _ = [self.destroy_subscription(_sub) for _sub in self.subs]\n",
    "        self.destroy_node()\n",
    "\n",
    "\n",
    "localize_node = LocalizeNode()\n",
    "nav_node = NavigateNode()\n",
    "\n",
    "def get_vlam_pose():\n",
    "    try:\n",
    "        rclpy.spin(localize_node)\n",
    "    except KeyboardInterrupt:\n",
    "        pass\n",
    "    #node.destroy()\n",
    "    return localize_node.pose\n",
    "\n",
    "def set_pos_absolute(x,y,deg, vxy=0.2, vw=0.5, tx=0.1, ty=0.1, tr=5.0, force=1):\n",
    "    try:\n",
    "        nav_node.set_gain()\n",
    "        rclpy.spin(nav_node)\n",
    "    except KeyboardInterrupt:\n",
    "        pass\n",
    "\n",
    "    try:\n",
    "        nav_node.run_pos(x,y,deg, vxy, vw, tx, ty, tr, force)\n",
    "        rclpy.spin(nav_node)\n",
    "    except KeyboardInterrupt:\n",
    "        pass\n",
    "\n",
    "\n",
    "def join():\n",
    "    try:\n",
    "        rclpy.spin(nav_node)\n",
    "    except KeyboardInterrupt:\n",
    "        pass\n",
    "\n",
    "def force_stop():\n",
    "    try:\n",
    "        nav_node.stop()\n",
    "        rclpy.spin(nav_node)\n",
    "    except KeyboardInterrupt:\n",
    "        pass\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7b8af8fe-a7bd-4480-9204-d0dac9e0d321",
   "metadata": {},
   "source": [
    "### 現在位置確認\n",
    "Visual SLAMにより推定された姿勢を取得します.  単位はx[<span style=\"color:#FF0000\">vm</span>], y[<span style=\"color:#FF0000\">vm</span>], z[<span style=\"color:#FF0000\">vm</span>], roll[rad], pitch[rad], yaw[rad], deg[°]です.<br>\n",
    "実際に使用する値は, x,y,deg の3つです. degはyawの値を[°]に変換したものです.<br>\n",
    "x,y,zの単位が[vm]となっているのは, スケールvが地図ごとに0.25~2.0程度変動するためです. vは地図作成終了まで不定です.<br>\n",
    "また, 現在位置は停止状態でも光の加減によって値が微小に変化します.  "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2d557e59-1ea2-4b79-80b1-6b271f8a8438",
   "metadata": {},
   "outputs": [],
   "source": [
    "pose = get_vlam_pose()\n",
    "print(\"x :\", pose.x)\n",
    "print(\"y :\", pose.y)\n",
    "print(\"z :\", pose.z)\n",
    "print(\"roll :\", pose.roll)\n",
    "print(\"pitch:\", pose.pitch)\n",
    "print(\"yaw  :\", pose.yaw)\n",
    "print(\"x,y,w = ({}, {}, {})\".format(pose.x, pose.y, pose.deg) )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "41e2a950-aca5-4ec0-8635-83a2cb569bba",
   "metadata": {},
   "source": [
    "## 絶対位値指令\n",
    "地図上のx,y,deg座標に移動します. 単位は, x[vm], y[vm], deg[°], vxy[m/s], vw[rad/s] です. <br>\n",
    "force=1の場合, 位置フィードバック制御を行います. このとき, 目標姿勢と現在姿勢の差がtx[vm], ty[vm], tr[°]になるか, 自己位置を見失うまで移動し続けます. <br>\n",
    "force=0を指定すると位置フィードバック制御を行わず, スケールを考慮しないため, 指示と異なる姿勢で止まることが多くなります."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "adea5e81-108f-431f-9f9e-5e4df9794892",
   "metadata": {},
   "outputs": [],
   "source": [
    "# -x方向（左）に 0.2[m/s] で移動. 停止精度をx=0.01[vm], y=0.01[vm], w=3[°] とする.\n",
    "pose = get_vlam_pose()\n",
    "print(\"現在姿勢 : \", pose.x,pose.y,pose.deg)\n",
    "# 目標姿勢\n",
    "x   = pose.x-0.2\n",
    "y   = pose.y\n",
    "deg = pose.deg\n",
    "# 移動速度\n",
    "vxy=0.2\n",
    "vw=0.5\n",
    "# 停止精度\n",
    "tx=0.01\n",
    "ty=0.01\n",
    "tr=3.0\n",
    "# 位置フィードバック\n",
    "force=1\n",
    "\n",
    "# 移動指示\n",
    "set_pos_absolute(x,y,deg, vxy, vw, tx, ty, tr, force)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6e1f5004-a3e1-4baa-83bc-6312597f5552",
   "metadata": {},
   "outputs": [],
   "source": [
    "pose = get_vlam_pose()\n",
    "print(\"移動後姿勢 : \", pose.x, pose.y, pose.deg)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7816231a",
   "metadata": {},
   "source": [
    "## 移動の中断"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "id": "2f4ccf27-0101-4116-9752-fa4a5fd49fb1",
   "metadata": {},
   "outputs": [],
   "source": [
    "# 移動を止めたい場合に実行\n",
    "force_stop()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "343c103d-8509-494b-a5ce-a65f8b56bb80",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
